#lang racket

(define (make-markers)
  (for/vector ([i 10])
    (let ([v (make-vector 10 #t)])
      (vector-set! v 0 #f)  
      v)))

(define (array-set! array r c v)
  (vector-set! (vector-ref array r) c v))

(define (array-ref array r c)
  (vector-ref (vector-ref array r) c))

(define (getBlk r c) ; block from row and column
  (+ (* 3 (quotient r 3)) (quotient c 3)))

(define sudoku%
  (class object%

    (init [puzzle-string ""])
    
    (define avail-row (make-markers))
    (define avail-col (make-markers))
    (define avail-blk (make-markers))
    (define count 0)
    
    (define grid
      (for/vector ([i 9]) (make-vector 9 0)))
    
    (super-new)

    (define/public (item-set! r c n)
      (array-set! grid r c n)
      (array-set! avail-row r n #f)
      (array-set! avail-col c n #f)
      (let ([b (getBlk r c)])
        (array-set! avail-blk b n #f))
      (set! count (+ count 1)))

    (define/public (get-grid) grid)
    
    (define/public (item-ref r c)
      (array-ref grid r c))

    (define/public (init-grid grid)
      (for* ([r 9] [c 9])
        (let ([n (array-ref grid r c)])
          (when (> n 0)
            (item-set! r c n)))))
    
    (define/private (init-puzzle p)
      (let ([g 
             (let ([rows (string-split p)])
               (for/vector ([row rows])
                 (for/vector ([c 9])
                   (string->number (substring row c (add1 c))))))])
        (init-grid g)))

    (define/public (print)
      (for* ([r 9] [c 9])
        (when (zero? c) (printf "\n"))
        (let ([n (item-ref r c)])
          (if (zero? n)
              (printf " .")
              (printf " ~a" n)
              )))
      (printf "\n"))

    (define/public (solved?) (= count 81))

    (define/public (clone)
      (let ([p (new sudoku%)])
        (send p init-grid grid)
        p))

    (define (avail r c)
      (let* ([b (getBlk r c)]
             [ar (vector-ref avail-row r)]
             [ac (vector-ref avail-col c)]
             [ab (vector-ref avail-blk b)])
        (for/vector ([i 10])
          (and (vector-ref ar i)
               (vector-ref ac i)
               (vector-ref ab i)))))

    (define (free-numbers v)
      (for/list ([n (in-range 1 10)] #:when (vector-ref v n)) n))
       
    (define (set-singles)
      (let ([found #f])
        (for* ([r 9]
               [c 9])
          (let* ([free (avail r c)]
                 [num-free (vector-count identity free)]
                 [n (item-ref r c)])
            (when (and (zero? n)  (= 1 num-free))
              (let ([first-free
                     (let loop ([i 1])
                       (if (vector-ref free i) i
                           (loop (add1 i))))])
                (item-set! r c first-free)
                (set! found #t))
              )))
        found))           

    (define/public (set-all-singles)
      (when (set-singles) (set-all-singles)))

    (define (get-free)
      (let ([free-list '()])
        (for* ([r 9]
               [c 9])
          (let* ([free (avail r c)]
                 [num-free (vector-count identity free)]
                 [n (item-ref r c)])
            (when (zero? n)
              (set! free-list
                    (cons
                     (list r c num-free (free-numbers free))
                     free-list)))))
        free-list))
    
    (define/public (get-min-free)
      (let ([min-free 10]
            [min-info null]
            [free-list (get-free)])
        (let loop ([free free-list])
          (unless (equal? free '())
            (let* ([info (car free)]
                   [rem (cdr free)]
                   [num-free (third info)])
              (when (< 0 num-free min-free)
                (set! min-free num-free)
                (set! min-info info))
              (loop rem))))
        min-info))

    (unless (equal? puzzle-string "")
      (init-puzzle puzzle-string))
    ))

(define (solve-sudoku puzzle)
  (let ([solution null]
        [puzzle (send puzzle clone)])
    (define (dfs puzzle)
      (if (send puzzle solved?)
          (set! solution puzzle)
          (let ([info (send puzzle get-min-free)]) 
            (match info
              ['() #f]
              [(list row col num free-nums) 
               (let loop ([nums free-nums])
                 (if (equal? nums '())
                     #f
                     (let ([n (car nums)]
                           [t (cdr nums)])
                       (let ([p (send puzzle clone)])
                         (send p item-set! row col n)
                         (send p set-all-singles)
                         (unless (dfs p)(loop t))))))]))))
    (send puzzle set-all-singles)
    (dfs puzzle)
    (if (equal? solution null)
        (error "No solution found.")
        solution
        )))

#;(define (solve pstr)
  (let* ([puzzle (new sudoku% [puzzle-string pstr])]
         [solution (solve-sudoku puzzle)])
    (send puzzle print)
    (send solution print)))

(define pstr "
150000320
604031097
000200000
000049073
400800000
000000001
503008000
002000004
007000260
")

;[define puzzle (new sudoku% [puzzle-string pstr])]

(require racket/draw)
(define CELL-SIZE 30)

(define (draw-centered-text dc text x y)
  (let-values ([(w h d s) (send dc get-text-extent text)])
    (let ([x (+ x (/ (- CELL-SIZE w) 2))]
          [y (+ y (/ (- CELL-SIZE h d) 2))])
      (send dc draw-text text x y ))))

(define (draw-puzzle p1 p2)
  (let* ([drawing (make-bitmap (* 9 CELL-SIZE) (* 9 CELL-SIZE))]
         [dc (new bitmap-dc% [bitmap drawing])]
         [yellow (new brush% [color (make-object color% 240 210 0)])]
         [gray (new brush% [color "Gainsboro"])])
    (for* ([r 9][c 9])
      (let* ([x (* c CELL-SIZE)]
             [y (* r CELL-SIZE)]
             [n1 (send p1 item-ref r c)]
             [n2 (send p2 item-ref r c)]
             [num (if (zero? n2) "" (number->string n2))]
             [color (if (zero? n1) yellow gray)])

                     
        (send dc set-pen "black" 1 'solid)
        (send dc set-brush color)
        (send dc draw-rectangle x y CELL-SIZE CELL-SIZE)
        (draw-centered-text dc num x y)))
    (for* ([r 3][c 3])
      (let* ([x (* 3 c CELL-SIZE)]
             [y (* 3 r CELL-SIZE)])
                     
        (send dc set-pen "black" 3 'solid)
        (send dc set-brush "black" 'transparent)
        (send dc draw-rectangle x y (* 3 CELL-SIZE) (* 3 CELL-SIZE))))
    drawing))

(define (solve pstr)
  (let* ([puzzle (new sudoku% [puzzle-string pstr])]
         [solution (solve-sudoku puzzle)])
    (print (draw-puzzle puzzle puzzle))
    (newline)
    (newline)
    (print (draw-puzzle puzzle solution))))

(solve "
150000320
604031097
000200000
000049073
400800000
000000001
503008000
002000004
007000260
")